from chemmltoolkit.tensorflow.metrics.metrics_mean_nan \
    import MeanAbsoluteErrorNaN, MeanSquaredErrorNaN, \
    RootMeanSquaredErrorNaN

__all__ = ['MeanAbsoluteErrorNaN', 'MeanSquaredErrorNaN',
           'RootMeanSquaredErrorNaN']



def get_feature_keys(feature, feature_length=1):
    if hasattr(feature, 'get_feature_keys'):
        return feature.get_feature_keys()
    else:
        feature_name = feature.__name__
        if feature_length > 1:
            return [f'{feature_name}[{i}]' for i in range(feature_length)]
        else:
            return [feature_name]


def get_token_name(token):
    return str(token)



from collections.abc import Iterable
from collections.abc import MutableMapping


def flatten(l: list) -> list:  # noqa: E741
    """Flattens a nested list into a flat list of elements.

    Args:
        l: The list to flatten.

    Returns:
        The flattened list.
    """
    def _flatten(l):  # noqa: E741
        for el in l:
            if _is_iterable(el):
                yield from flatten(el)
            else:
                yield el

    return list(_flatten(l))


def flatten_dict(d: dict, separator: str = '.') -> dict:
    """Flattens a nested dictionary into a flat dictionary.

    This will flatten any nested dictionary items using the specified
    separator, for example if the separator is '.' then in the form
    'grandparent.parent.key'.

    Args:
        d: The dictionary to flatten.
        separator: The string to separate levels of the dictionary.

    Returns:
        The flattened dictionary.
    """
    def _flatten_dict(d, separator, prefix):
        items = []
        for k, v in d.items():
            new_key = prefix + k
            if isinstance(v, MutableMapping):
                new_prefix = new_key + separator
                items.extend(_flatten_dict(v, separator, new_prefix))
            else:
                items.append((new_key, v))
        return items

    items = _flatten_dict(d, separator, '')
    return dict(items)


def merge_dict(d: dict, defaults: dict) -> dict:
    """Merges a dictionary with a set of default values.

    This will combine the entries in two dictionaries, with the first
    argument taking preference.

    Args:
        d: The dictionary of specified values.
        default: The dictionary of default values.

    Returns:
        The merged dictionary.
    """
    return {**defaults, **d}


def one_hot(feature, tokens: list) -> list:
    """One-hot encodes a feature.

    This will one-hot encode the feature, using a specified list of tokens.

    Args:
        feature: The feature value to one-hot encode.
        tokens: A list of tokens to use for encoding.

    Returns:
        A one-hot encoded list.
    """
    return [int(feature == token) for token in tokens]


def pad_list(l: list, length: int, item) -> list:  # noqa: E741
    """Pads a list to the specified length.

    Note that if the input list is longer than the specified length it is not
    truncated.

    Args:
        l: The input list.
        length: The desired length of the list.
        item: The item to use to pad the list.

    Returns:
        The padded list.
    """
    return l + [item] * (length - len(l))


def zip_expand(*args):
    """Zips the arguments, expanding them to the largest length if required.

    This will perform a zip of the specified arguments. The resulting list
    will be of the same length as the longest input. If any arguments are
    not lists then they will be repeated to the required length.

    Args:
        args: The elements to zip.

    Returns:
        The zipped list.
    """
    def _len_arg(arg):
        return len(arg) if _is_iterable(arg) else 1

    def _expand_arg(arg, size):
        return arg if _is_iterable(arg) else [arg] * size

    size = max([_len_arg(arg) for arg in args])
    args = [_expand_arg(arg, size) for arg in args]
    return zip(*args)


def _is_iterable(x):
    return isinstance(x, Iterable) and not isinstance(x, (str, bytes))



from chemmltoolkit.tensorflow.layers.preprocessing.addSelfLoops \
    import AddSelfLoops
from chemmltoolkit.tensorflow.layers.preprocessing.normaliseGraph \
    import NormaliseGraph

__all__ = ['AddSelfLoops', 'NormaliseGraph']



from chemmltoolkit.utils.list_utils import pad_list


def desalt_smiles(smiles: str) -> str:
    """Removes all salts from the specified SMILES strings.

    This is a very simple and quick desalting algorithm that returns
    the longest component withing the input string. This will fail
    for cases with embedded components (e.g. 'c1cc([O-].[Na+])cccc1')
    or where individual components are actually connected
    (e.g. ethane represented as 'C1.C1').

    Args:
        smiles: The input SMILES string.

    Returns:
        The desalted SMILES string.
    """
    components = smiles.split('.')
    return max(components, key=len)


def split_smiles(smiles: str) -> list:
    """Splits a SMILES string into individual tokens.

    This will split a SMILES string into tokens representing
    individual parts of the SMILES string. Each individual token
    will be one of the following,

    - A single atom (including hydrogens or charge if specified)
    - A ring connection number (single digit or multi-digit)
    - Another SMILES character

    Args:
        smiles: The input SMILES string.

    Returns:
        A list of tokens.
    """
    result = []
    while smiles != '':
        if smiles[:2] == 'Cl' or smiles[:2] == 'Br':
            token_length = 2
        elif smiles[0] == '%':  # Ring linkage numbers >9 are prefixed by '%'
            token_length = 3
        elif smiles[0] == '[':
            token_length = smiles.index(']') + 1
        else:
            token_length = 1

        result.append(smiles[:token_length])
        smiles = smiles[token_length:]
    return result


def _split_smiles_halogen_only(smiles: str) -> list:
    result = []
    while smiles != '':
        if smiles[:2] == 'Cl' or smiles[:2] == 'Br':
            token_length = 2
        else:
            token_length = 1

        result.append(smiles[:token_length])
        smiles = smiles[token_length:]
    return result


def get_token_list(max_ring_number=9, chiral_carbon=True) -> list:
    tokens = []

    # SMILES 'organic' subset that do not need '[...]'
    tokens += ['B', 'C', 'N', 'O', 'P', 'S', 'F', 'Cl', 'Br', 'I']
    tokens += ['c', 'n', 'o', 's']
    tokens += ['[nH]', '[n+]', '[N+]', '[O-]']

    # Chiral atoms
    if chiral_carbon:
        tokens += ['[C@]', '[C@@]', '[C@H]', '[C@@H]']

    # Control tokens
    tokens += [
        '-', '=', '#',  # Bonds
        '(', ')',  # Atom groups
        '/', '\\',  # Double bond geometry
    ]

    # Ring tokens
    tokens += [f'{x}' for x in range(1, min(max_ring_number, 9) + 1)] \
        + [f'%{x}' for x in range(10, max_ring_number + 1)]

    return tokens


class SmilesTokeniser(object):
    """Tokeniser to convert SMILES strings to and from tokenised format.

    Before tokenising, the SMILES string is split into its constituent
    components. This can be performed in a number of ways
        - 'all_tokens' will correctly split all SMILES tokens
        - 'halogens_only' is character based apart from Cl and Br
        - 'characters' is entirely character based

    Args:
        tokens: A list of strings containing the token list to use.
        splitting_method: The technique for splitting the SMILES string.
        token_unknown: The token to use for unidentified tokens, or
            None if SMILES string with unknown tokens should be rejected.
        token_sos: A token to add at the start of the sequence.
        token_eos: A token to add at the end of the sequence.
        token_padding: A token to use to pad to the desired length.
        sequence_length: The desired length of the resulting sequence.
        truncate_sequence: Whether to truncate the sequence if it is longer
            than 'sequence_length'.
        simplify_rings: Will re-use ring number tokens where possible.
    """

    def __init__(self,
                 tokens: list,
                 splitting_method: str = 'all_tokens',
                 token_unknown: str = None,
                 token_sos: str = None,
                 token_eos: str = None,
                 token_padding: str = None,
                 sequence_length: int = None,
                 truncate_sequence: bool = False,
                 simplify_rings: bool = False):

        if (token_padding and not sequence_length) or \
           (sequence_length and not token_padding):
            raise ValueError('Both `token_padding` and `sequence_length`' +
                             'must be set together.')

        self._tokens_list = tokens
        self._tokens_missing = set()
        self._token_lookup = {t: i for i, t in enumerate(tokens)}
        self._token_unknown = tokens.index(token_unknown)\
            if token_unknown else None
        self._sequence_sos = [tokens.index(token_sos)]\
            if token_sos else []
        self._sequence_eos = [tokens.index(token_eos)]\
            if token_eos else []
        self._token_padding = tokens.index(token_padding)\
            if token_padding else None
        self._sequence_length = sequence_length
        self._truncate_sequence = truncate_sequence
        self._splitting_method = splitting_method
        self._simplify_rings = simplify_rings

        if splitting_method == 'all_tokens':
            self._split_function = split_smiles
        elif splitting_method == 'halogens_only':
            self._split_function = _split_smiles_halogen_only
        elif splitting_method == 'characters':
            self._split_function = list

    @property
    def missing_tokens(self):
        """Returns a set containing any SMILES tokens that have not been
        found in the supplied tokens list.
        """
        return self._tokens_missing

    def tokenise_smiles(self, smiles: str) -> list:
        """Tokenises the specified SMILES string.

        This function will take a SMILES string and return the corresponding
        tokens as a list of integer values. The processing used is
        specified when creating this SmilesTokeniser.

        Args:
            smiles: The SMILES string to tokenise.
            unknown_placeholder: The token to use for unidentified tokens, or
                None if SMILES string with unknown tokens should be rejected.

        Returns:
            A list of integer tokens.
        """

        # If we are not splitting by 'all_tokens' and ring numbering is greater
        # than 9 (i.e. %xx numbering) fallback to pre-simplifying ring with
        # 'all_tokens' first, then use the desired splitting method
        preprocess_ring_simplification = self._simplify_rings \
            and self._splitting_method != 'all_tokens' \
            and '%' in smiles

        if preprocess_ring_simplification:
            components = split_smiles(smiles)
            components = self._ring_simplifier(components)
            smiles = ''.join(components)

        components = self._split_function(smiles)

        if self._simplify_rings and not preprocess_ring_simplification:
            components = self._ring_simplifier(components)

        smiles_tokens = [self._get_token(c) for c in components]

        if None in smiles_tokens:
            return None

        smiles_tokens = self._sequence_sos + smiles_tokens + self._sequence_eos

        if self._token_padding:
            smiles_tokens = pad_list(smiles_tokens,
                                     self._sequence_length,
                                     self._token_padding)

            if self._truncate_sequence:
                smiles_tokens = smiles_tokens[:self._sequence_length]

        return smiles_tokens

    def untokenise_smiles(self, smiles_tokens: list) -> str:
        """Returns the SMILES string corresponding to the tokenised representation.

        This function will take a list of tokens and will return the decoded
        SMILES string. The processing used is
        specified when creating this SmilesTokeniser.

        Args:
            smiles_tokens: A list of integer tokens.

        Returns:
            The corresponding SMILES string.
        """
        components = [self._tokens_list[t] for t in smiles_tokens]

        if self._simplify_rings:
            # If we aren't using 'all_tokens' then swap to this format
            if self._splitting_method != 'all_tokens':
                components = split_smiles(''.join(components))
            components = self._ring_unsimplifier(components)

        return ''.join(components)

    def _get_token(self, val):
        token = self._token_lookup.get(val)
        if token is not None:
            return token
        else:
            self._tokens_missing.add(val)
            return self._token_unknown

    def _ring_simplifier(self, components):
        ring_rename_dictionary = {}
        num_tokens = ['1', '2', '3', '4', '5', '6', '7', '8', '9', '%']

        def get_next_id():
            for id in range(1, 100):
                id_str = str(id) if id < 10 else '%' + str(id)
                if id_str not in ring_rename_dictionary:
                    return id_str

        def simplify_component(component):
            if component[0] in num_tokens:
                if component not in ring_rename_dictionary:
                    new_id = get_next_id()
                    ring_rename_dictionary[component] = new_id
                    return new_id
                else:
                    return ring_rename_dictionary.pop(component)
            else:
                return component

        return [simplify_component(c) for c in components]

    def _ring_unsimplifier(self, components):
        ring_rename_dictionary = {}
        num_tokens = ['1', '2', '3', '4', '5', '6', '7', '8', '9', '%']
        next_id = 1

        def unsimplify_component(component):
            nonlocal next_id
            if component[0] in num_tokens:
                if component not in ring_rename_dictionary:
                    id = str(next_id) if next_id < 10 else '%' + str(next_id)
                    next_id += 1
                    ring_rename_dictionary[component] = id
                    return id
                else:
                    return ring_rename_dictionary.pop(component)
            else:
                return component

        return [unsimplify_component(c) for c in components]



import os
import shutil
import urllib3
import certifi
import tarfile
import zipfile

http = urllib3.PoolManager(cert_reqs='CERT_REQUIRED', ca_certs=certifi.where())


def get_file(filename,
             url,
             cache_subdir='downloads',
             cache_dir=None):

    if cache_dir is None:
        cache_dir = os.path.join(os.path.expanduser('~'), '.chemmltoolkit')

    datadir_base = os.path.expanduser(cache_dir)
    datadir = os.path.join(datadir_base, cache_subdir)
    fpath = os.path.join(datadir, filename)

    if not os.path.exists(datadir):
        os.makedirs(datadir)

    if not os.path.exists(fpath):
        print('Downloading data from', url)

        with http.request('GET', url, preload_content=False) as r, \
                open(fpath, 'wb') as out_file:
            shutil.copyfileobj(r, out_file)

    return fpath


def extract_all(archive,
                cache_subdir=None,
                cache_dir=None):

    if cache_dir is None:
        cache_dir = os.path.join(os.path.expanduser('~'), '.chemmltoolkit')

    if cache_subdir is None:
        archive_name = os.path.basename(archive).split(os.path.extsep)[0]
        cache_subdir = os.path.join('downloads', archive_name)

    datadir_base = os.path.expanduser(cache_dir)
    datadir = os.path.join(datadir_base, cache_subdir)

    if not os.path.exists(datadir):
        print('Extracting datafile', archive)
        os.makedirs(datadir)

        if archive.endswith(".tar.bz2"):
            with tarfile.open(archive, "r:bz2") as tar:
                tar.extractall(path=datadir)
        elif archive.endswith(".zip"):
            with zipfile.ZipFile(archive) as zfile:
                zfile.extractall(path=datadir)

    return datadir



import numpy as np
from rdkit.Chem import Mol
from rdkit.Chem import DataStructs
from rdkit.Chem import rdFingerprintGenerator
import rdkit.Chem.Descriptors as Desc


def _fingerprint_fn_bits(generator):
    def _fp(mol: Mol):
        fingerprint = generator.GetFingerprint(mol)
        array = np.zeros((0,), dtype=np.int8)
        DataStructs.ConvertToNumpyArray(fingerprint, array)
        return array

    return _fp


def _fingerprint_fn_count(generator):
    def _fp(mol: Mol):
        fingerprint = generator.GetCountFingerprint(mol)
        array = np.zeros((0,), dtype=np.int8)
        DataStructs.ConvertToNumpyArray(fingerprint, array)
        return array

    return _fp


def all_rdkit(exception_list=[]):
    """Gets a list of all RDKit descriptors.

    This will return a set of features for all descriptors in the
    rdkit.Chem.Descriptors.descList property. To allow repeatability, this
    list is hardcoded from RDKit 2019.09.3.

    """
    descriptors_2019_09_3 = [
        'MaxEStateIndex', 'MinEStateIndex', 'MaxAbsEStateIndex',
        'MinAbsEStateIndex', 'qed', 'MolWt', 'HeavyAtomMolWt', 'ExactMolWt',
        'NumValenceElectrons', 'NumRadicalElectrons', 'MaxPartialCharge',
        'MinPartialCharge', 'MaxAbsPartialCharge', 'MinAbsPartialCharge',
        'FpDensityMorgan1', 'FpDensityMorgan2', 'FpDensityMorgan3', 'BalabanJ',
        'BertzCT', 'Chi0', 'Chi0n', 'Chi0v', 'Chi1', 'Chi1n', 'Chi1v', 'Chi2n',
        'Chi2v', 'Chi3n', 'Chi3v', 'Chi4n', 'Chi4v', 'HallKierAlpha', 'Ipc',
        'Kappa1', 'Kappa2', 'Kappa3', 'LabuteASA', 'PEOE_VSA1', 'PEOE_VSA10',
        'PEOE_VSA11', 'PEOE_VSA12', 'PEOE_VSA13', 'PEOE_VSA14', 'PEOE_VSA2',
        'PEOE_VSA3', 'PEOE_VSA4', 'PEOE_VSA5', 'PEOE_VSA6', 'PEOE_VSA7',
        'PEOE_VSA8', 'PEOE_VSA9', 'SMR_VSA1', 'SMR_VSA10', 'SMR_VSA2',
        'SMR_VSA3', 'SMR_VSA4', 'SMR_VSA5', 'SMR_VSA6', 'SMR_VSA7', 'SMR_VSA8',
        'SMR_VSA9', 'SlogP_VSA1', 'SlogP_VSA10', 'SlogP_VSA11', 'SlogP_VSA12',
        'SlogP_VSA2', 'SlogP_VSA3', 'SlogP_VSA4', 'SlogP_VSA5', 'SlogP_VSA6',
        'SlogP_VSA7', 'SlogP_VSA8', 'SlogP_VSA9', 'TPSA', 'EState_VSA1',
        'EState_VSA10', 'EState_VSA11', 'EState_VSA2', 'EState_VSA3',
        'EState_VSA4', 'EState_VSA5', 'EState_VSA6', 'EState_VSA7',
        'EState_VSA8', 'EState_VSA9', 'VSA_EState1', 'VSA_EState10',
        'VSA_EState2', 'VSA_EState3', 'VSA_EState4', 'VSA_EState5',
        'VSA_EState6', 'VSA_EState7', 'VSA_EState8', 'VSA_EState9',
        'FractionCSP3', 'HeavyAtomCount', 'NHOHCount', 'NOCount',
        'NumAliphaticCarbocycles', 'NumAliphaticHeterocycles',
        'NumAliphaticRings', 'NumAromaticCarbocycles',
        'NumAromaticHeterocycles', 'NumAromaticRings', 'NumHAcceptors',
        'NumHDonors', 'NumHeteroatoms', 'NumRotatableBonds',
        'NumSaturatedCarbocycles', 'NumSaturatedHeterocycles',
        'NumSaturatedRings', 'RingCount', 'MolLogP', 'MolMR', 'fr_Al_COO',
        'fr_Al_OH', 'fr_Al_OH_noTert', 'fr_ArN', 'fr_Ar_COO', 'fr_Ar_N',
        'fr_Ar_NH', 'fr_Ar_OH', 'fr_COO', 'fr_COO2', 'fr_C_O', 'fr_C_O_noCOO',
        'fr_C_S', 'fr_HOCCN', 'fr_Imine', 'fr_NH0', 'fr_NH1', 'fr_NH2',
        'fr_N_O', 'fr_Ndealkylation1', 'fr_Ndealkylation2', 'fr_Nhpyrrole',
        'fr_SH', 'fr_aldehyde', 'fr_alkyl_carbamate', 'fr_alkyl_halide',
        'fr_allylic_oxid', 'fr_amide', 'fr_amidine', 'fr_aniline',
        'fr_aryl_methyl', 'fr_azide', 'fr_azo', 'fr_barbitur', 'fr_benzene',
        'fr_benzodiazepine', 'fr_bicyclic', 'fr_diazo', 'fr_dihydropyridine',
        'fr_epoxide', 'fr_ester', 'fr_ether', 'fr_furan', 'fr_guanido',
        'fr_halogen', 'fr_hdrzine', 'fr_hdrzone', 'fr_imidazole', 'fr_imide',
        'fr_isocyan', 'fr_isothiocyan', 'fr_ketone', 'fr_ketone_Topliss',
        'fr_lactam', 'fr_lactone', 'fr_methoxy', 'fr_morpholine', 'fr_nitrile',
        'fr_nitro', 'fr_nitro_arom', 'fr_nitro_arom_nonortho', 'fr_nitroso',
        'fr_oxazole', 'fr_oxime', 'fr_para_hydroxylation', 'fr_phenol',
        'fr_phenol_noOrthoHbond', 'fr_phos_acid', 'fr_phos_ester',
        'fr_piperdine', 'fr_piperzine', 'fr_priamide', 'fr_prisulfonamd',
        'fr_pyridine', 'fr_quatN', 'fr_sulfide', 'fr_sulfonamd', 'fr_sulfone',
        'fr_term_acetylene', 'fr_tetrazole', 'fr_thiazole', 'fr_thiocyan',
        'fr_thiophene', 'fr_unbrch_alkane', 'fr_urea']

    return [rdkit(descriptor) for descriptor in descriptors_2019_09_3
            if descriptor not in exception_list]


def fingerprint_atompair(fpSize=2048, count=False):
    """Atom pair fingerprint (list of int).

    Args:
        fpSize: Size of the generated fingerprint (defaults to 2048).
        count: The default value of False will generate fingerprint bits
            (0 or 1) whereas a value of True will generate the count of each
            fingerprint value.
    """
    generator = rdFingerprintGenerator.GetAtomPairGenerator(fpSize=fpSize)

    if count:
        fingerprint_fn = _fingerprint_fn_count(generator)
    else:
        fingerprint_fn = _fingerprint_fn_bits(generator)

    fingerprint_fn.__name__ = 'fingerprint_atompair(' + \
                              f'fpSize={fpSize},count={count})'
    return fingerprint_fn


def fingerprint_morgan(radius, fpSize=2048, count=False):
    """Morgan fingerprint of the specified size (list of int).

    Args:
        radius: The number of iterations to grow the fingerprint.
        fpSize: Size of the generated fingerprint (defaults to 2048).
        count: The default value of False will generate fingerprint bits
            (0 or 1) whereas a value of True will generate the count of each
            fingerprint value.
    """
    generator = rdFingerprintGenerator.GetMorganGenerator(radius=radius,
                                                          fpSize=fpSize)

    if count:
        fingerprint_fn = _fingerprint_fn_count(generator)
    else:
        fingerprint_fn = _fingerprint_fn_bits(generator)

    fingerprint_fn.__name__ = f'fingerprint_morgan(radius={radius},' + \
                              f'fpSize={fpSize},count={count})'
    return fingerprint_fn


def logp(mol: Mol) -> float:
    """Calculated LogP (float).
    """
    return Desc.MolLogP(mol)


def molwt(mol: Mol) -> float:
    """Molecular weight (float).
    """
    return Desc.MolWt(mol)


def num_atoms(mol: Mol) -> int:
    """Total number of atoms (int).
    """
    return mol.GetNumAtoms()


def num_bonds(mol: Mol) -> int:
    """Total number of bonds (int).
    """
    return mol.GetNumBonds()


def num_h_donors(mol: Mol) -> int:
    """Number of hydrogen bond donors (int).
    """
    return Desc.NumHDonors(mol)


def num_h_acceptors(mol: Mol) -> int:
    """Number of hydrogen bond acceptors (int).
    """
    return Desc.NumHAcceptors(mol)


def num_heavy_atoms(mol: Mol) -> int:
    """Number of heavy atoms (int).
    """
    return Desc.HeavyAtomCount(mol)


def rdkit(name: str):
    """Feature from a specified RDKit descriptor.

    The descriptor should be the name for the corresponding descriptor
    in rdkit.Chem.Descriptors.

    Args:
        name: The name of the descriptor.
    """
    descriptor = getattr(Desc, name)

    def _rdkit(mol: Mol):
        return descriptor(mol)
    _rdkit.__name__ = f'rdkit({name})'
    return _rdkit


def tpsa(mol: Mol) -> float:
    """Total polar surface area (float)
    """
    return Desc.TPSA(mol)



from typing import List
from chemmltoolkit.features.featuriser import Featuriser
import numpy as np
from rdkit.Chem import MolFromSmiles
from rdkit.Chem import Mol
from rdkit.Chem import Bond


class BondFeaturiser(Featuriser):
    """Generator for bond-based features.

    Args:
        features: A list of features to generate.
    """
    def __init__(self, features: list):
        super(BondFeaturiser, self).__init__(features)

    def process_bond(self, bond: Bond):
        """Generates features for an individual bond.

        Args:
            bond: The bond to featurise.

        Returns:
            A list of features.
        """
        return self._process(bond)

    def process_molecule(self, mol: Mol):
        """Generates features for all bond in a molecule.

        Args:
            mol: The molecule to featurise.

        Returns:
            A list of tuples for all bonds. Each tuple contains
                (from_atom_index, to_atom_index, features)
        """
        bonds = mol.GetBonds()
        return [(
            bond.GetBeginAtomIdx(),
            bond.GetEndAtomIdx(),
            self._process(bond)) for bond in bonds]

    def generate_adjacency_matricies(self, mol: Mol):
        """Generates adjacency matricies for a molecule.

        Args:
            mol: The molecule to featurise.

        Returns:
            A list of adjacency matricies, each for an individual feature
        """
        num_atoms = mol.GetNumAtoms()
        bond_features = self.process_molecule(mol)
        num_features = len(bond_features[0][2])

        adj = np.zeros((num_features, num_atoms, num_atoms))

        for atom_begin, atom_end, features in bond_features:
            adj[:, atom_begin, atom_end] = features
            adj[:, atom_end, atom_begin] = features

        return adj

    def get_feature_lengths(self) -> List[int]:
        """Calculates the length of each feature

        Returns:
            A list of the lengths of each feature.
        """
        molecule = MolFromSmiles('CC')
        bond = molecule.GetBonds()[0]
        return self._get_feature_lengths(bond)



import tensorflow as tf
from tensorflow.python.framework import dtypes
from tensorflow.python.framework import tensor_shape
from tensorflow.python.ops import nn
from tensorflow.python.keras import activations
from tensorflow.python.keras import initializers
from tensorflow.python.keras import regularizers
from tensorflow.python.keras import constraints
from tensorflow.keras.layers import Layer
from tensorflow.keras.layers import InputSpec
import tensorflow.keras.backend as K
from chemmltoolkit.tensorflow.graph.tensorGraph import NODE_FEATURES
from chemmltoolkit.tensorflow.graph.tensorGraph import EDGE_FEATURES
from chemmltoolkit.tensorflow.utils import register_keras_custom_object


@register_keras_custom_object
class GraphConv(Layer):
    """A graph convolutional layer.
    Arguments:
        units: Positive integer, dimensionality of the output space.
        activation: Activation function to use.
            If you don't specify anything, RELU activation is applied.
        use_bias: Boolean, whether the layer uses a bias vector.
        add_self_loops: Boolean, whether to add a separate weight matrix
            for self loops.
        aggregation_method: Specifies the method to use when combining outputs
            if multiple adjacency matricies are specified ('concat', 'sum' or
            'max').
        graph_regularization: The type of graph regularization to use. The
            default (None) is to not use any regularization. Possible values
            are 'basis' (basis decomposition).
        num_bases: The number of bases to use for graph regularization. This
            must be set if graph_regularization is used, otherwise left as the
            default (None).
        kernel_initializer: Initializer for the `kernel` weights matrix.
        kernel_coef_initializer: Initializer for the `kernel` coefficients
            matrix (only used with graph regularization).
        bias_initializer: Initializer for the bias vector.
        kernel_regularizer: Regularizer function applied to
            the `kernel` weights matrix.
        kernel_coef_regularizer: Regularizer function applied to
            the `kernel` coefficients matrix (only used with graph
            regularization).
        bias_regularizer: Regularizer function applied to the bias vector.
        activity_regularizer: Regularizer function applied to
            the output of the layer (its "activation")..
        kernel_constraint: Constraint function applied to
            the `kernel` weights matrix.
        kernel_constraint: Constraint function applied to
            the `kernel` coefficients matrix (only used with graph
            regularization).
        bias_constraint: Constraint function applied to the bias vector.
    Input shape:
        Two tensors are provided as input.
        An N-D feature tensor with shape:
            `(batch_size, ..., input_dim)`.
        A 3-D adjacency tensor with shape:
            `(batch_size, num_adjacency, input_dim, input_dim)`.
    Output shape:
        N-D tensor with shape: `(batch_size, ..., units)`.
        For instance, for a 2D input with shape `(batch_size, input_dim)`,
        the output would have shape `(batch_size, units)`.
    """

    def __init__(self,
                 units,
                 activation='relu',
                 use_bias=True,
                 add_self_loops=False,
                 aggregation_method='sum',
                 graph_regularization=None,
                 num_bases=None,
                 kernel_initializer='glorot_uniform',
                 kernel_coef_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 kernel_regularizer=None,
                 kernel_coef_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 kernel_coef_constraint=None,
                 bias_constraint=None,
                 **kwargs):
        super(GraphConv, self).__init__(
            activity_regularizer=regularizers.get(activity_regularizer),
            **kwargs)

        if graph_regularization not in [None, 'basis']:
            raise ValueError('An invalid value of `graph_regularization` has' +
                             ' been passed to a `GraphConv` layer.')

        if graph_regularization is not None and num_bases is None:
            raise ValueError('The `num_bases` property must be set if ' +
                             '`graph_regularization` is not None.')

        if graph_regularization is None and num_bases is not None:
            raise ValueError('The `num_bases` property must not be set if ' +
                             '`graph_regularization` is None.')

        if num_bases is not None and num_bases <= 0:
            raise ValueError('The `num_bases` property must be a positive ' +
                             'integer.')

        self.units = int(units)
        self.activation = activations.get(activation)
        self.use_bias = use_bias
        self.add_self_loops = add_self_loops
        self.aggregation_method = aggregation_method
        self.graph_regularization = graph_regularization
        self.num_bases = num_bases
        self.kernel_initializer = initializers.get(kernel_initializer)
        self.kernel_coef_initializer = initializers.get(
            kernel_coef_initializer)
        self.bias_initializer = initializers.get(bias_initializer)
        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        self.kernel_coef_regularizer = regularizers.get(
            kernel_coef_regularizer)
        self.bias_regularizer = regularizers.get(bias_regularizer)
        self.kernel_constraint = constraints.get(kernel_constraint)
        self.kernel_coef_constraint = constraints.get(kernel_coef_constraint)
        self.bias_constraint = constraints.get(bias_constraint)

        self.input_spec = {
            NODE_FEATURES: InputSpec(min_ndim=2),
            EDGE_FEATURES: InputSpec(ndim=4)
        }

    def build(self, input_shape):
        # Validate inputs
        if not isinstance(input_shape, dict):
            raise ValueError('The `GraphConv` layer expects a dict of inputs.')
        if NODE_FEATURES not in input_shape:
            raise ValueError('A `GraphConv` layer should be called with a '
                             'dict containing key "' + NODE_FEATURES + '".')
        if EDGE_FEATURES not in input_shape:
            raise ValueError('A `GraphConv` layer should be called with a '
                             'dict containing key "' + EDGE_FEATURES + '".')

        dtype = dtypes.as_dtype(self.dtype or K.floatx())
        if not (dtype.is_floating or dtype.is_complex):
            raise TypeError('Unable to build `GraphConv` layer with '
                            'non-floating point dtype %s' % (dtype,))

        features_shape = tensor_shape.TensorShape(input_shape[NODE_FEATURES])
        adjacency_shape = tensor_shape.TensorShape(input_shape[EDGE_FEATURES])
        if tensor_shape.dimension_value(features_shape[-1]) is None:
            raise ValueError('The last dimension of the features matrix '
                             '(first input) to `GraphConv` should be defined.'
                             'Found `None`.')
        if tensor_shape.dimension_value(features_shape[-2]) != \
                tensor_shape.dimension_value(adjacency_shape[-1]) \
                or tensor_shape.dimension_value(features_shape[-2]) != \
                tensor_shape.dimension_value(adjacency_shape[-2]):
            raise ValueError('The last two dimensions of the adjacency matrix'
                             '(second input) to `GraphConv` should be the '
                             'same as the penultimate dimension of the '
                             'features matrix (first input). '
                             f'Found {adjacency_shape} for feature size ' +
                             tensor_shape.dimension_value(features_shape[-2]))

        if self.num_bases \
                and self.num_bases > \
                tensor_shape.dimension_value(adjacency_shape[1]):
            raise ValueError('The adjacency matrix (second input) passed to ' +
                             '`GraphConv` must not have less adjacency ' +
                             'matricies than `num_bases`.')

        # Define the input spec
        self.num_node_features = tensor_shape.dimension_value(
            features_shape[-1])
        self.num_nodes = tensor_shape.dimension_value(adjacency_shape[2])
        self.num_edge_features = tensor_shape.dimension_value(
            adjacency_shape[1])

        self.input_spec = {
            NODE_FEATURES: InputSpec(min_ndim=2, axes={
                -1: self.num_node_features,
                -2: self.num_nodes}),
            EDGE_FEATURES: InputSpec(ndim=4, axes={
                1: self.num_edge_features,
                2: self.num_nodes,
                3: self.num_nodes})
        }

        # Add weights
        num_self_loops = 1 if self.add_self_loops else 0
        self.num_weights = self.num_edge_features + num_self_loops
        num_kernels = self.num_bases if self.graph_regularization else \
            self.num_weights

        self.kernel = self.add_weight(
            'kernel',
            shape=[num_kernels,
                   self.num_node_features,
                   self.units],
            initializer=self.kernel_initializer,
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint,
            dtype=self.dtype,
            trainable=True)

        if self.graph_regularization:
            self.kernel_coef = self.add_weight(
                'kernel_coef',
                shape=[self.num_weights,
                       self.num_bases],
                initializer=self.kernel_coef_initializer,
                regularizer=self.kernel_coef_regularizer,
                constraint=self.kernel_coef_constraint,
                dtype=self.dtype,
                trainable=True)

        if self.use_bias:
            self.bias = self.add_weight(
                'bias',
                shape=[self.units, ],
                initializer=self.bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
                dtype=self.dtype,
                trainable=True)
        else:
            self.bias = None

        self.built = True

    def call(self, inputs):
        features = tf.convert_to_tensor(inputs[NODE_FEATURES])
        is_sparse_adjacency = isinstance(inputs[EDGE_FEATURES],
                                         tf.sparse.SparseTensor)
        if is_sparse_adjacency:
            adjacency = tf.sparse.to_dense(inputs[EDGE_FEATURES])
        else:
            adjacency = tf.convert_to_tensor(inputs[EDGE_FEATURES])

        weights = self._get_weight_matrix()

        features = tf.expand_dims(features, 1)
        outputs = tf.matmul(adjacency, features)

        if self.add_self_loops:
            outputs = tf.concat([features, outputs], 1)

        outputs = tf.matmul(outputs, weights)

        if self.aggregation_method == 'concat':
            outputs = tf.transpose(outputs, (0, 2, 1, 3))
            outputs_shape = tf.shape(outputs)
            outputs = tf.reshape(outputs,
                                 (outputs_shape[0], outputs_shape[1], -1))
        elif self.aggregation_method == 'sum':
            outputs = tf.reduce_sum(outputs, axis=1)
        elif self.aggregation_method == 'max':
            outputs = tf.reduce_max(outputs, axis=1)
        else:
            raise ValueError('Undefined aggregation method' +
                             f'`{self.aggregation_method}`.')

        if self.use_bias:
            outputs = nn.bias_add(outputs, self.bias)
        if self.activation is not None:
            outputs = self.activation(outputs)  # pylint: disable=not-callable

        return {**inputs, NODE_FEATURES: outputs}

    def _get_weight_matrix(self):
        # Get the weight matrix by applying the specified regularization
        # method (if any) to the kernel matrix
        if self.graph_regularization == 'basis':
            weights = tf.reshape(self.kernel, (self.num_bases,
                                               self.num_node_features *
                                               self.units))
            weights = tf.matmul(self.kernel_coef, weights)
            weights = tf.reshape(weights, (self.num_weights,
                                           self.num_node_features, self.units))
            return weights
        else:
            return self.kernel

    def compute_output_shape(self, input_shape):
        features_shape = tensor_shape.TensorShape(input_shape[NODE_FEATURES])
        adjacency_shape = tensor_shape.TensorShape(input_shape[EDGE_FEATURES])
        features_shape = features_shape.with_rank_at_least(2)

        if self.aggregation_method == 'concat':
            multiplier = tensor_shape.dimension_value(adjacency_shape[1])
        else:
            multiplier = 1

        out_shape = features_shape[:-1].concatenate(self.units * multiplier)

        return {
            NODE_FEATURES: out_shape,
            EDGE_FEATURES: adjacency_shape
        }

    def get_config(self):
        config = {
            'units': self.units,
            'activation': activations.serialize(self.activation),
            'use_bias': self.use_bias,
            'add_self_loops': self.add_self_loops,
            'aggregation_method': self.aggregation_method,
            'graph_regularization': self.graph_regularization,
            'num_bases': self.num_bases,
            'kernel_initializer':
                initializers.serialize(self.kernel_initializer),
            'kernel_coef_initializer':
                initializers.serialize(self.kernel_coef_initializer),
            'bias_initializer': initializers.serialize(self.bias_initializer),
            'kernel_regularizer':
                regularizers.serialize(self.kernel_regularizer),
            'kernel_coef_regularizer':
                regularizers.serialize(self.kernel_coef_regularizer),
            'bias_regularizer': regularizers.serialize(self.bias_regularizer),
            'activity_regularizer':
                regularizers.serialize(self.activity_regularizer),
            'kernel_constraint': constraints.serialize(self.kernel_constraint),
            'kernel_coef_constraint':
                constraints.serialize(self.kernel_coef_constraint),
            'bias_constraint': constraints.serialize(self.bias_constraint)
        }
        base_config = super(GraphConv, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



import tensorflow as tf
from tensorflow.keras.layers.experimental.preprocessing \
    import PreprocessingLayer
from chemmltoolkit.tensorflow.graph.tensorGraph import map_edge_features
from chemmltoolkit.tensorflow.utils import register_keras_custom_object


@register_keras_custom_object
class AddSelfLoops(PreprocessingLayer):
    """A preprocessing layer to add self-loops to TensorGraphs.

    Self-loops are additional connections between each node and itself where
    the edge features are all one.

    Input shape:
        A TensorGraph is provided as input.
    Output shape:
        A TensorGraph with identical shape to the input.
    """

    def __init__(self, name=None, **kwargs):
        super(AddSelfLoops, self).__init__(name=name, **kwargs)

    def call(self, inputs):
        def _call(edge_features):
            eye = tf.eye(tf.shape(edge_features)[-1])
            return edge_features + eye

        return map_edge_features(inputs, _call)

    def compute_output_shape(self, input_shape):
        return input_shape

    def get_config(self):
        config = super(AddSelfLoops, self).get_config()
        return config



import tensorflow as tf
from tensorflow.keras.layers.experimental.preprocessing \
    import PreprocessingLayer
from chemmltoolkit.tensorflow.graph.tensorGraph import map_edge_features
from chemmltoolkit.tensorflow.utils import register_keras_custom_object


@register_keras_custom_object
class NormaliseGraph(PreprocessingLayer):
    """A preprocessing layer to normalise TensorGraphs.
    Arguments:
        method: The method of normalisation to use. Possible options are,
            'normal' (default): Scales all rows to sum to one
            'spectral': Uses the spectral method of Kipf and Welling
                (https://arxiv.org/abs/1609.02907)
    Input shape:
        A TensorGraph is provided as input.
    Output shape:
        A TensorGraph with identical shape to the input.
    """

    def __init__(self, name=None, method='normal', **kwargs):
        super(NormaliseGraph, self).__init__(name=name, **kwargs)
        self.method = method

    def call(self, inputs):
        if self.method == 'spectral':
            map_fn = self.normalise_spectral
        else:
            map_fn = self.normalise_normal

        return map_edge_features(inputs, map_fn)

    def normalise_normal(self, edge_features):
        degree = tf.reduce_sum(edge_features, axis=-2)
        degree = tf.expand_dims(degree, -1)
        return tf.math.divide_no_nan(edge_features, degree)

    def normalise_spectral(self, edge_features):
        degree = tf.reduce_sum(edge_features, axis=-2)
        degree_inv = tf.pow(degree, -0.5)
        degree_inv = tf.where(tf.math.is_inf(degree_inv), 0.0, degree_inv)

        # (degree_inv)^T . matrix . (degree_inv)
        di1 = tf.expand_dims(degree_inv, -2)
        di2 = tf.expand_dims(degree_inv, -1)
        return tf.multiply(tf.multiply(di1, edge_features), di2)

    def compute_output_shape(self, input_shape):
        return input_shape

    def get_config(self):
        config = super(NormaliseGraph, self).get_config()
        config.update({'method': self.method})
        return config



import tensorflow as tf


# @tf.keras.utils.register_keras_serializable(package='ChemMLToolkit')
@tf.function
def mean_squared_error_nan(y_true, y_pred):
    """Computes the mean squared error, masking any NaN values in `y_true`.
    Arguments:
        y_true: Tensor of true targets (possibly including NaN values).
        y_pred: Tensor of predicted targets.
    Returns:
        Tensor with one scalar loss entry per sample.
    """
    y_pred = tf.convert_to_tensor(y_pred)
    y_true = tf.cast(y_true, y_pred.dtype)

    mask = ~tf.math.is_nan(y_true)
    y_true = tf.where(mask, y_true, 0.0)
    y_pred = tf.where(mask, y_pred, 0.0)

    error = tf.reduce_sum((y_pred - y_true) ** 2, axis=-1)
    count = tf.reduce_sum(tf.cast(mask, y_pred.dtype), axis=-1)
    loss = tf.math.divide_no_nan(error, count)
    return loss


mse_nan = MSE_nan = mean_squared_error_nan


# @tf.keras.utils.register_keras_serializable(package='ChemMLToolkit')
class MeanSquaredErrorNaN(tf.keras.losses.Loss):
    """Computes the mean squared error, masking any NaN values in `y_true`.
    Arguments:
        reduction: (Optional) Type of loss reduction to apply to loss. The
            default value is `SUM_OVER_BATCH_SIZE`.
        name: (Optional) name for the loss.
    """

    def __init__(self,
                 reduction=tf.keras.losses.Reduction.SUM_OVER_BATCH_SIZE,
                 name="mean_squared_error_nan_loss"):
        super(MeanSquaredErrorNaN, self).__init__(
            reduction=reduction, name=name)

    def call(self, y_true, y_pred):
        return mean_squared_error_nan(y_true, y_pred)



from chemmltoolkit.features.decorators import tokenizable_feature
from rdkit.Chem import Atom
from rdkit.Chem import AllChem
from rdkit.Chem import ChiralType
from rdkit.Chem import HybridizationType


def atomic_number(atom: Atom) -> int:
    """Atomic number (int).
    """
    return atom.GetAtomicNum()


def atomic_mass(atom: Atom) -> float:
    """Atomic mass (float).
    """
    return atom.GetMass()


def charge(atom: Atom) -> int:
    """Formal charge (int).
    """
    return atom.GetFormalCharge()


def charge_gasteiger(atom: Atom) -> float:
    """Gasteiger partial charge (float).
    """
    if not atom.HasProp('_GasteigerCharge'):
        mol = atom.GetOwningMol()
        AllChem.ComputeGasteigerCharges(mol)
    return atom.GetDoubleProp('_GasteigerCharge')


def charge_gasteiger_h(atom: Atom) -> float:
    """Gasteiger partial charge for implicit hydrogens (float).
    """
    if not atom.HasProp('_GasteigerHCharge'):
        mol = atom.GetOwningMol()
        AllChem.ComputeGasteigerCharges(mol)
    return atom.GetDoubleProp('_GasteigerHCharge')


@tokenizable_feature([ChiralType.CHI_UNSPECIFIED,
                      ChiralType.CHI_TETRAHEDRAL_CW,
                      ChiralType.CHI_TETRAHEDRAL_CCW,
                      ChiralType.CHI_OTHER])
def chiral_tag(atom: Atom) -> ChiralType:
    """Chirality of the atom (ChiralType)
    """
    return atom.GetChiralTag()


def degree(atom: Atom) -> int:
    """Number of directly bonded neighbours (int).
    """
    return atom.GetDegree()


@tokenizable_feature([HybridizationType.SP,
                      HybridizationType.SP2,
                      HybridizationType.SP3,
                      HybridizationType.SP3D,
                      HybridizationType.SP3D2])
def hybridization(atom: Atom) -> HybridizationType:
    """Hybridisation (HybridizationType).
    """
    return atom.GetHybridization()


def hydrogens(atom: Atom) -> int:
    """Total number of hydrogen atoms (int).
    """
    return atom.GetTotalNumHs()


def index(atom: Atom) -> int:
    """Index within the parent molecule (int).
    """
    return atom.GetIdx()


def is_aromatic(atom: Atom) -> int:
    """If the atom is aromatic (0 or 1).
    """
    return int(atom.GetIsAromatic())


def is_ring(atom: Atom) -> int:
    """If the atom is is in a ring (0 or 1).
    """
    return int(atom.IsInRing())


def is_ringsize(ringSize: int) -> int:
    """If the atom is is in a ring of the specified size (0 or 1).

    Args:
        ringSize: The size of the ring.
    """
    def _is_ringsize(atom: Atom):
        return int(atom.IsInRingSize(ringSize))
    _is_ringsize.__name__ = f'is_ringsize({ringSize})'
    return _is_ringsize


def isotope(atom: Atom) -> int:
    """Isotope (int).
    """
    return atom.GetIsotope()


def radical(atom: Atom) -> int:
    """Number of radical electrons (int).
    """
    return atom.GetNumRadicalElectrons()


def symbol(atom: Atom) -> str:
    """Atomic symbol (string).
    """
    return atom.GetSymbol()



from typing import Tuple
import re

from chemmltoolkit.data.quantities import Quantity
from chemmltoolkit.data.quantities._constants \
    import unit_prefix_dict, units


class RegexCache():
    def __init__(self):
        unit_prefix_regex = '|'.join(unit_prefix_dict.keys())
        units_regex = '|'.join(units)
        units_full_regex = r'((?P<unit_prefix>' + unit_prefix_regex + ')' + \
            r'(?P<units>' + units_regex + '))?'

        operator_regex = r'(?P<operator>(=|~|<|>|<=|>=)?)'
        value_regex = r'(?P<value>[-+]?[\d\.]+)'
        min_value_regex = value_regex.replace('value', 'min_value')
        max_value_regex = value_regex.replace('value', 'max_value')

        self.quantity = re.compile(
            rf'^{operator_regex}{value_regex}$')

        self.range = re.compile(
            rf'^{min_value_regex}-{max_value_regex}$')

        self.quantity_with_units = re.compile(
            rf'^{operator_regex}{value_regex}\s*{units_full_regex}$')

        self.range_with_units = re.compile(
            rf'^{min_value_regex}-{max_value_regex}' +
            rf'\s*{units_full_regex}$')


_regex_cache = None


def from_string(input: str, return_units: bool = False) -> Quantity:
    global _regex_cache
    if not _regex_cache:
        _regex_cache = RegexCache()

    match = _regex_cache.quantity.match(input)
    if match:
        operator = match.group('operator')
        value = float(match.group('value'))

        operator = None if operator in ['', '='] else operator
        quantity = Quantity.from_value(value, operator)
        return quantity

    match = _regex_cache.range.match(input)
    if match:
        min_value = float(match.group('min_value'))
        max_value = float(match.group('max_value'))

        quantity = Quantity(min_value, max_value, False, False, 0.0)
        return quantity

    raise ValueError(f"Input is not a valid Quantity: '{input}'")


def from_string_with_units(input: str) -> Tuple[Quantity, str]:
    global _regex_cache
    if not _regex_cache:
        _regex_cache = RegexCache()

    match = _regex_cache.quantity_with_units.match(input)
    if match:
        operator = match.group('operator')
        value = float(match.group('value'))
        unit_prefix = match.group('unit_prefix')
        units = match.group('units')

        operator = None if operator in ['', '='] else operator
        exponent = unit_prefix_dict[unit_prefix] if unit_prefix else 1.0
        value = value * exponent

        quantity = Quantity.from_value(value, operator)
        return (quantity, units)

    match = _regex_cache.range_with_units.match(input)
    if match:
        min_value = float(match.group('min_value'))
        max_value = float(match.group('max_value'))
        unit_prefix = match.group('unit_prefix')
        units = match.group('units')

        exponent = unit_prefix_dict[unit_prefix] if unit_prefix else 1.0
        min_value = min_value * exponent
        max_value = max_value * exponent

        quantity = Quantity(min_value, max_value, False, False, 0.0)
        return (quantity, units)

    raise ValueError(f"Input is not a valid Quantity: '{input}'")


def to_string(q: Quantity) -> str:
    if not q.operator:
        return f'{q.value}'
    elif isinstance(q.value, float):
        return f'{q.operator}{q.value}'
    else:
        min_value, max_value = q.value
        return f'{min_value}{q.operator}{max_value}'



from chemmltoolkit.tensorflow.graph.tensorGraph import TensorGraph
from chemmltoolkit.tensorflow.graph.tensorGraph import TensorGraphShape
from chemmltoolkit.tensorflow.graph.tensorGraph import map_node_features
from chemmltoolkit.tensorflow.graph.tensorGraph import map_edge_features
from chemmltoolkit.tensorflow.graph.tensorGraphInput import TensorGraphInput

__all__ = ['TensorGraph', 'TensorGraphShape', 'map_node_features',
           'map_edge_features', 'TensorGraphInput']



from __future__ import annotations
from typing import Any, Optional, Tuple, Union
import numpy as np


class Quantity:
    def __init__(self, val_min: float, val_max: float,
                 eq_min: bool, eq_max: bool, error: float):
        if val_min > val_max:
            raise ValueError(
                '`val_min` must be less than or equal to `val_max`.')

        self.val_min = val_min
        self.val_max = val_max
        self.eq_min = eq_min
        self.eq_max = eq_max
        self.error = error

    @property
    def value(self) -> Union[float, Tuple[float, float]]:
        if self.val_min == self.val_max:
            return self.val_min
        if self.val_min == -np.inf:
            return self.val_max
        elif self.val_max == np.inf:
            return self.val_min
        else:
            return (self.val_min, self.val_max)

    @property
    def operator(self) -> Optional[str]:
        if self.val_min == self.val_max:
            return '~' if self.error == np.inf else None
        if self.val_min == -np.inf:
            return '<=' if self.eq_max else '<'
        elif self.val_max == np.inf:
            return '>=' if self.eq_min else '>'
        else:
            return '-'

    @staticmethod
    def from_value(value: Union[float, Tuple[float, float]],
                   operator: Optional[str]):
        if operator == '-':
            if not isinstance(value, tuple) or (len(value) != 2) \
                    or (type(value[0]) is not float) \
                    or (type(value[1]) is not float):
                raise ValueError('The `value` must be a tuple of two floats.')
            return Quantity(value[0], value[1], True, True, 0.0)
        else:
            if type(value) is not float:
                raise ValueError('The `value` must be a float.')

            if operator is None:
                return Quantity(value, value, True, True, 0.0)
            elif operator == '~':
                return Quantity(value, value, True, True, np.inf)
            elif operator == '>':
                return Quantity(value, np.inf, False, False, 0.0)
            elif operator == '<':
                return Quantity(-np.inf, value, False, False, 0.0)
            elif operator == '>=':
                return Quantity(value, np.inf, True, False, 0.0)
            elif operator == '<=':
                return Quantity(-np.inf, value, False, True, 0.0)
            else:
                raise ValueError(f'Invalid `operator` "{operator}".')

    def __str__(self):
        from chemmltoolkit.data.quantities.string_conv import to_string
        return to_string(self)

    def __repr__(self):
        from chemmltoolkit.data.quantities.string_conv import to_string
        return to_string(self)

    def __add__(self, other: Quantity):
        from chemmltoolkit.data.quantities.maths import add
        return add(self, other)

    def __eq__(self, other: Any):
        if not isinstance(other, Quantity):
            return False

        return self.val_min == other.val_min \
            and self.val_max == other.val_max \
            and self.eq_min == other.eq_min \
            and self.eq_max == other.eq_max \
            and self.error == other.error

    def __neg__(self):
        from chemmltoolkit.data.quantities.maths import neg
        return neg(self)

    def __pos__(self):
        return Quantity(self.val_min, self.val_max,
                        self.eq_min, self.eq_max, self.error)

    def __sub__(self, other: Quantity):
        from chemmltoolkit.data.quantities.maths import sub
        return sub(self, other)



from chemmltoolkit.features.atomFeaturiser import AtomFeaturiser
from chemmltoolkit.features.bondFeaturiser import BondFeaturiser
from chemmltoolkit.features.moleculeFeaturiser import MoleculeFeaturiser

__all__ = ['AtomFeaturiser', 'BondFeaturiser', 'MoleculeFeaturiser']
